/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "minrtsFftUtil.h"
#include "delay.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#define SMOOTHSIGMA  0.4 /* 400 ms default */
#define AMPSMOOTH   16.0 /* seconds */
#define DELAYSMOOTH 10.0 /* points smoothing in delay direction */
#define MEDSMOOTH   15.0 /* seconds median smoothing before output*/
#define ENDDROP 0.1  /* fraction pts at ends of delayogram to drop */
/*#define DEBUG 1*/
void Usage()
{
  printf ("Usage: oirPowerDelay GroupDelayFile OutputDelayFile [-smooth smooth] [-ampsmooth ampsmooth]\n");
  printf ("      [-delaysmooth delaysmooth] [-medsmooth medsmooth] \n");
  printf("    smooth:1 sigma width of gaussian coherent smoothing interval(sec)\n");
  printf("        same usage as in oirFGroupDelay; default=0.4sec\n");
  printf("    ampsmooth:box car smoothing interval (sec) of abs(delay)^2\n");
  printf("        after above coherent smoothing; default=16sec\n");
  printf("    delaysmooth:1 sigma width of gaussian smooth in delay direction(pixels~1micron)\n");
  printf("        after above ampsmooth; default=10pixels\n");
  printf("    medsmooth:median smoothing interval (sec) of of computed OPD peaks default=15sec\n\n");
  exit(1);
}
/* ************************************************************************
 * Find the index and value of the maximum and mininum of a float array
 * Inputs:
 *         a    array of input values 
 *         n    number of input values
 * Outputs:
 *         *min value of mininum of array
 *         *max value of maximum of array
 *         *imin position (0-rel) of minimum value in array
 *         *imax position (0-rel) of maximum value in array
 *   
 *********************************************************************** */
void mmax(float *a, int n, float *min, float *max, int *imin, int *imax) {
   int i,j,k;
   float in,ax;
   j=0;
   k=0;
   in = a[0];
   ax = in;
   for (i=0;i<n;i++){
      if (a[i] > ax) {
         ax = a[i];
         j  = i;
      } 
      if (a[i] < in) {
         in = a[i];
         k  = i;
      } 
   }
   *imin = k;
   *imax = j;
   *min  = in;
   *max  = ax;
   return;
}

/*
 *     add a new value (in) to a sorted list (list[nList])
*/    
void addSortedD(double *list, int nList, double in) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL &&(in > list[i])) i++;
   for (j=nL;j>i;j--) list[j] = list[j-1];
   list[i] = in;
   return;
}
/*
 *     remove a  value (out) to a sorted list (list[nList])
*/    
void removeSortedD(double *list, int nList, double out) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL&&(out != list[i])) i++;
   for(j=i;j<nL;j++) list[j] = list[j+1];
   return;
}
/*
 *     combine the previous two operations
*/    
void addRemoveSortedD(double *list, int nList, double in, double out) {
   int iout, iin, i;
   if (out == in) return;
   iout=0;
   while (iout < nList &&(out != list[iout])) iout++;
   if (in < out) { 
      iin = 0;
      while (iin < iout &&(in > list[iin])) iin++;
      for (i=iout;i>iin;i--) list[i] = list[i-1];
   } else {
      iin = iout;
      while (iin < (nList-1) &&(in > list[iin+1])) iin++;
      for (i=iout;i<iin;i++) list[i] = list[i+1];
   }
   list[iin] = in;
   return;
}
void minrtsMedianSmoothD(double *data, int nData, int nSmooth, double* outData) {
   double *mData;
   int nOdd, n2;
   int i,nList, l2;
   int out, odd;

   odd = nSmooth - 2 * (nSmooth/2); /* I can only take the median of an odd number of pts */
   if (odd) nOdd = nSmooth; else nOdd = nSmooth+1;
   n2 = (int)(nOdd/2);
   mData = (double *)malloc(nOdd * sizeof(double)); /* work array to store current samples */

   for (i=0;i<=n2;i++) addSortedD(mData,i+1,data[i]); /* initial load of mData */
   nList = n2;
   for (out=0; out<n2;out++) {
      l2           = (int)(nList/2);
      nList++;
      outData[out] = mData[l2];
      addSortedD(mData, nList+1, data[nList]);
   }
   for (out=n2; out<(nData-n2-1); out++) {
      outData[out] = mData[n2];
      addRemoveSortedD(mData, nOdd, data[out+n2+1], data[out-n2]);
   }
   outData[nData-n2-1] = mData[n2];
   nList = n2;
   for (out=nData-n2; out<nData; out++) {
      removeSortedD(mData, n2+nList+1, data[nData-nList-2]);
      l2           = (int)(nList/2);
      outData[out] = mData[l2];
      nList--;
   }
   free(mData); 
   return;
}
/* **************************************************************************
 *  Smooth the complex frames in place with a boxcar of width smooth
 *  INPUTS:
 *         data                 input and output complex (2 Reals) data array 
 *         nData                number of points
 *         smooth               size of smoothing box
 *         work                 pre allocated buffer equal in size to data
 ************************************************************************** */
void complexBoxSmooth (float *data, int nData,  int smooth, float* work) {
   float rbox, ibox;
   int nbox, first, next;
   int i;

   rbox = 0.;
   ibox = 0.;
/*           load real/imaginary accumulators */
   for (i=0;i<(smooth-1);i++) {rbox=rbox+data[2*i]; ibox=ibox+data[2*i+1];}
   first = 0;
   next  = smooth-2;
   nbox  = smooth-1;
   for (i=0; i<nData;i++) {
      if (next < (nData-1)) {
         next++;
         rbox = rbox+data[2*next];
         ibox = ibox+data[2*next+1];
         nbox++;
      }
      if (first < (i-smooth)) {
         rbox = rbox - data[2*first];
         ibox = ibox - data[2*first+1];
         first++;
         nbox--;
      }
      work[2*i]   = rbox/nbox;
      work[2*i+1] = ibox/nbox;
   }
   memcpy(data, work, 2*nData*sizeof(float));
   return;
}

/* **************************************************************************
 *  Smooth the complex frames in place with a gaussian of given sigma 
 *  INPUTS:
 *         data                 input and output complex (2 Reals) data array 
 *         nData                number of points
 *         sigma                size of smoothing box
 *         work                 pre allocated buffer equal in size to data
 ************************************************************************** */
void complexGaussSmooth (float *data, int nData,  float sigma, float* work) {
   float *gauss;
   float  w,weight;
   int    iSigma;
   float  rl, img;
   int i, iG, imin, imax;

   iSigma = (int)(2*sigma);
   gauss  = (float*)malloc((2*iSigma+1)*sizeof(float));
   weight = 0.;
   for (i=-iSigma;i<iSigma;i++) 
      gauss[i+iSigma] = exp(-0.5*i*i/(sigma*sigma));

   for (i=0;i<nData;i++) {
      rl    = 0.;
      img   = 0.;
      weight= 0.;
      imin  = i-iSigma; if (imin<0) imin = 0; 
      imax  = i+iSigma; if (imax > nData) imax = nData;
      for (iG=imin;iG<imax;iG++) {
         w      = gauss[iG-i+iSigma];
         rl     = rl +data[2*iG]*w;
         img    = img+data[2*iG+1]*w;
         weight = weight+w;
      }
      work[2*i]   = rl/weight;
      work[2*i+1] = img/weight;
   }
   memcpy(data, work, 2*nData*sizeof(float));
   free(gauss);
   return;
}

/*
 *     add a new value (in) to a sorted list (list[nList])
*/    
void addSorted(float *list, int nList, float in) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL &&(in > list[i])) i++;
   for (j=nL;j>i;j--) list[j] = list[j-1];
   list[i] = in;
   return;
}
/*
 *     remove a  value (out) to a sorted list (list[nList])
*/    
void removeSorted(float *list, int nList, float out) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL&&(out != list[i])) i++;
   for(j=i;j<nL;j++) list[j] = list[j+1];
   return;
}
/*
 *     combine the previous two operations
*/    
void addRemoveSorted(float *list, int nList, float in, float out) {
   int iout, iin, i;
   if (out == in) return;
   iout=0;
   while (iout < nList &&(out != list[iout])) iout++;
   if (in < out) { 
      iin = 0;
      while (iin < iout &&(in > list[iin])) iin++;
      for (i=iout;i>iin;i--) list[i] = list[i-1];
   } else {
      iin = iout;
      while (iin < (nList-1) &&(in > list[iin+1])) iin++;
      for (i=iout;i<iin;i++) list[i] = list[i+1];
   }
   list[iin] = in;
   return;
}
void minrtsMedianSmooth(float *data, int nData, int nSmooth, float* outData) {
   float *mData;
   int nOdd, n2;
   int i,nList, l2;
   int out, odd;

   odd = nSmooth - 2 * (nSmooth/2); /* I can only take the median of an odd number of pts */
   if (odd) nOdd = nSmooth; else nOdd = nSmooth+1;
   n2 = (int)(nOdd/2);
   mData = (float *)malloc(nOdd * sizeof(float)); /* work array to store current samples */

   for (i=0;i<=n2;i++) addSorted(mData,i+1,data[i]); /* initial load of mData */
   nList = n2;
   for (out=0; out<n2;out++) {
      l2           = (int)(nList/2);
      nList++;
      outData[out] = mData[l2];
      addSorted(mData, nList+1, data[nList]);
   }
   for (out=n2; out<(nData-n2-1); out++) {
      outData[out] = mData[n2];
      addRemoveSorted(mData, nOdd, data[out+n2+1], data[out-n2]);
   }
   outData[nData-n2-1] = mData[n2];
   nList = n2;
   for (out=nData-n2; out<nData; out++) {
      removeSorted(mData, n2+nList+1, data[nData-nList-2]);
      l2           = (int)(nList/2);
      outData[out] = mData[l2];
      nList--;
   }
   free(mData); 
   return;
}
/* **************************************************************************
 *    compute absolute value of delay function, 
 *    find peak, do parabolic interpolation to subpixel accuracy
 *    around peak, and return position and value of peak
 *  INPUTS:
 *      fft        fft object containing coordinate and pixel information
 *      delayData  complex delay function data
 *  OUTPUTS:
 *      mx         value of interpolated peak of absolute value of delayData
 *  RETURNS:
 *                 position of interpolated peak of absolute value, in units specified by fft
 *      
 ************************************************************************** */
int main (int argc, char** argv) {

   int   i, j, ierr;
   float x;
   long  nRow, iRow, oRow, pRow ;
   int   nx, nx2, nx4 ;

   char  outFileName[130];
   char  *bang="!";
   char* infile_name     = NULL;
   char* outfile_name    = NULL;
   fitsfile *inFile;
   fitsfile *outFile = NULL;

   char history[80];

   int iKey, maxstep;
   double opd, opd0, opd1, opd2;
   int   nBegin, nEnd, nDrop, nSearch;

   imaging_fdata_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row **inData;
   imaging_fdata_row *outData;

   delay_table_row *delayData;
   delay_table_def *dTable;

   float *work1, *work2;


   float detdit;  /* exposure time to convert times->frames */
   float   smooth, ampSmooth, delaySmooth, medTime;
   int     medFrames;

   float amin, amax;
   int   imin, imax;
   float *delayCoord;
   double *powerDelays, *medianBuffer;

   int iarg;  /* for parsing input */

   minrtsFft1d *ff1[1], *ff2[1];
   ff1[0]   = NULL;
   ff2[0]   = NULL;
   ierr    = 0;


   /**
    * Parse the command line options 
    */
   if (argc < 3) {
      Usage();
      return 1;
   }
   infile_name  = argv[1];
   outfile_name = argv[2];

   smooth      = SMOOTHSIGMA; /* default */
   ampSmooth   = AMPSMOOTH;   /* default */
   delaySmooth = DELAYSMOOTH; /* default */
   medTime     = MEDSMOOTH;   /* default */
   iarg = 3; 
   while (iarg < argc ) {
      if (!strcmp("-smooth", argv[iarg])) sscanf(argv[++iarg],"%f",&smooth); 
      if (!strcmp("-ampsmooth", argv[iarg])) sscanf(argv[++iarg],"%f",&ampSmooth); 
      if (!strcmp("-delaysmooth", argv[iarg])) sscanf(argv[++iarg],"%f",&delaySmooth); 
      if (!strcmp("-medsmooth", argv[iarg])) sscanf(argv[++iarg],"%f",&medTime); 
      ++iarg;
   }
   printf("      ******   oirPowerDelay   ******\n");
/*
      open input file and get detector info
*/
   fits_open_file (&inFile, infile_name, READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i\n",infile_name,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &detdit,0,&ierr);
   printf("   exposure time (DIT)          = %8.3f sec\n", detdit);
   printf("   coherent smoothing sigma     = %8.3f sec ==%8.1f frames\n", smooth, smooth/detdit);
   printf("   amplitude smoothing sigma    = %8.3f sec ==%8.1f frames\n", ampSmooth, ampSmooth/detdit);
   smooth    = smooth/detdit;
   ampSmooth = ampSmooth/detdit;
   medFrames = medTime/detdit;
   printf("   median smoothing width       = %8.3f sec ==%6d   frames\n", medTime, medFrames);

/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, outfile_name);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits output %s %i\n",outfile_name, ierr);
      printf("%s \n",strerror(errno));
      return 1;
      }
/*
      copy header and add some new keywords
*/
   i = fits_copy_header(inFile, outFile, &ierr);
   snprintf(history,80,"oirGroupDelay INFILE %s", infile_name);
   i = fits_write_history(outFile, history, &ierr);
   i = fits_write_key(outFile, TFLOAT, "OSMOOTH", (void*)&smooth, 
      "Smoothing halfwidth in oirGroupDelay", &ierr);
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   nx  = DetectorDef->naxis[0][0]; /* pseudo complex */
   nx2 = nx/2; /* real length */
   nx4 = nx/4; /* half real length */
/*
      screw around with naxis; this is necessary
      because data in .insopd file is "pseudocomplex"
      so nx is twice what it is supposed to be.
*/
   DetectorDef->naxis[0][0] = nx2;
/*
      configure data table and get MAXSTEP
*/
   idata          = open_imaging_fdata_table(inFile, &ierr);
   iKey = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep,
      &ierr);
/*
 *      number of rows
*/ 
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   
   odata = make_imaging_fdata_table_def (outFile, 
      DetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: creating data definition\n");
      return 1;
      }
/*
         create enough space for entire input data
*/
   inData = (imaging_fdata_row **) malloc(nRow * sizeof(imaging_fdata_row *));
   for (iRow=0;iRow<nRow;iRow++) inData[iRow] = make_imaging_fdata_row(idata);
   outData = make_imaging_fdata_row(odata);

/*       
         get OPD offset value from header
*/
   iKey = table_find_key(idata->def, "OPD0");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&opd0, &ierr);
   iKey = table_find_key(idata->def, "OPD1");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&opd1, &ierr);
   iKey = table_find_key(idata->def, "OPD2");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&opd2, &ierr);

   printf("   smoothing delay over           %8.1f points  ==%8.0f microns\n",
      delaySmooth,-delaySmooth*opd2*1.e6);
   printf("   total number of frames       = %ld\n", nRow);
/*
         Read in input data
*/ 
   for (iRow=1; iRow<=nRow; iRow++) 
      i = read_imaging_fdata_table(inFile, idata, inData[iRow-1], iRow, &ierr);

/*
 *          alloc work arrays
*/
   work1 = (float *)malloc(2*nRow*sizeof(float));
   work2 = (float *)malloc(2*nRow*sizeof(float));
/*
 *          for each delay point in array smooth data in time direction
 *          first a complex boxcar smooth with size "smooth"
 *          then convert to amp^2
 *          then real median smooth by ampSmooth (standard deviation)
 *          then gaussian smooth in the delay direction by delaySmooth
 *          at the end return the data for the time being into the input array
*/
   for (i=0; i<nx2;i++) {
      for (iRow=0;iRow<(2*(int)smooth);iRow++) work1[iRow] = 0;
      for (iRow=(int)smooth;iRow<nRow;iRow++) {
         work1[2*iRow]   = inData[iRow]->data[0][2*i];
         work1[2*iRow+1] = inData[iRow]->data[0][2*i+1];
      }
/*     complexBoxSmooth(work1, nRow, smooth, work2);  */
      complexGaussSmooth(work1, nRow, smooth, work2); 
      for (iRow=0;iRow<nRow;iRow++) 
         work1[iRow] = work1[2*iRow]*work1[2*iRow]+work1[2*iRow+1]*work1[2*iRow+1]; /* abs val ^2 */
       if (ampSmooth < 0) minrtsGaussSmooth(-ampSmooth, nRow, work1, work2, ff1, 0); else
           minrtsBoxSmooth(nRow, work1, work2, ampSmooth);
       for (iRow=0;iRow<nRow;iRow++) inData[iRow]->data[0][i] = work2[iRow]; 
       j = i/50;
       if ((i-j*50)==0) fprintf(stderr,"Row: %i of %i\r",i, nx2);

   }
   if (ampSmooth < 0) minrtsKillFft1d(ff1[0]);
   fprintf(stderr,"");
/*           smooth in delay direction */
   for (iRow=0;iRow<nRow;iRow++) {
      for (i=0; i<nx2;i++) work1[i]   = inData[iRow]->data[0][i];
      minrtsGaussSmooth(delaySmooth, nx2, work1, inData[iRow]->data[0], ff2, 0);
   }
   minrtsKillFft1d(ff2[0]);
/*
 *          write output data
*/
   for (iRow=0;iRow<nRow;iRow++) {
/*
            first copy nonData info
*/

      outData->time      = inData[iRow]->time;
      outData->opd[0]    = inData[iRow]->opd[0];
      outData->opd[1]    = inData[iRow]->opd[1];
      outData->local_opd[1] = inData[iRow]->local_opd[1];
      outData->local_opd[0] = inData[iRow]->local_opd[0];
      outData->exp       = inData[iRow]->exp;
      outData->offset[0] = inData[iRow]->offset[0];
      outData->offset[1] = inData[iRow]->offset[1];
      outData->rotation  = inData[iRow]->rotation;
      outData->frame     = inData[iRow]->frame;
      outData->step      = inData[iRow]->step;
      outData->reference = inData[iRow]->reference;
      outData->targtype[0] = inData[iRow]->targtype[0];  
/*
         now copy data into output row
*/
      memcpy (outData->data[0], inData[iRow]->data[0], nx2*sizeof(float));
      i = write_imaging_fdata_table(outFile, odata, outData, iRow+1, &ierr);
   }

   i = fits_write_key(outFile, TDOUBLE, "OPD0", (void*)(&opd0), "OFFSET TO OPD", &ierr);
   i = fits_write_key(outFile, TDOUBLE,  "OPD1", (void*)(&opd1), " ", &ierr);
   i = fits_write_key(outFile, TDOUBLE,  "OPD2", (void*)(&opd2), " ", &ierr);
   kill_imaging_fdata_table_def(idata);
   kill_imaging_fdata_table_def(odata);
   kill_imaging_fdata_row(outData);


/*
       calculate and output estimates of power delay
*/
   dTable = make_delay_table_def(outFile, "1776-07-04", &ierr);
/*   i = fits_write_key(outFile, TFLOAT, "DNOISE", (void*)(&amplitudeRMS), 
      "Delay Amplitude Noise", &ierr);
*/
   delayData = make_delay_table_row(dTable);
   delayCoord = (float *)malloc(nx2*sizeof(float));
   for (i=0;i<nx2;i++) delayCoord[i] = opd0 + opd2*(i-nx4);
   nBegin = ENDDROP * nx2;
   nSearch = nx2-2*ENDDROP;
   medianBuffer = (double *)malloc(nRow*sizeof(double));
   powerDelays  = (double *)malloc(nRow*sizeof(double));
   for (iRow=0;iRow<nRow;iRow++) {
      mmax(inData[iRow]->data[0]+nBegin, nSearch, &amin, &amax, &imin, &imax);
      imax = imax + nBegin;
      if (imax < 3 || (imax > (nx2 - 4))) medianBuffer[iRow] = delayCoord[imax];
      else {
         x = minrtsQPk(7, delayCoord+imax-3, inData[iRow]->data[0]+imax-3, &amax, 0);
         if (x > delayCoord[imax-3]) x = delayCoord[imax-3];
         if (x < delayCoord[imax+3]) x = delayCoord[imax+3];
         medianBuffer[iRow] = x;
      }
   }
   minrtsMedianSmoothD(medianBuffer, nRow, medFrames, powerDelays);
   for (iRow=0;iRow<nRow;iRow++) {
      delayData->time      = inData[iRow]->time;
      delayData->telescope = 2;
      delayData->amplitude = sqrt(amax);
      delayData->delay     = powerDelays[iRow]/clight; /* in seconds */
      i = write_delay_table(outFile, dTable, delayData, iRow+1, &ierr);
   }
   free(powerDelays);
   kill_delay_table_def(dTable);
   kill_delay_table_row(delayData);
   free (delayCoord);

   fits_close_file(inFile, &ierr);
   fits_close_file(outFile, &ierr);
   free(work1);
   free(work2);
   for (iRow=0;iRow<nRow;iRow++) kill_imaging_fdata_row(inData[iRow]);
   free(inData);

   return 0;
}
